from flask import current_app, request
from flask_login import current_user, login_required
from . import socketio
from .decorators import admin_required
from .models import User
import json
import jsonpatch
import logging


'''
' A list containing session ids of connected Socket.IO sessions, to keep track
' of all connected sessions, which is used to determine the runtimes of
' associated background tasks.
'''
connected_sessions = []


@socketio.on('connect')
@login_required
def connect():
    '''
    ' The Socket.IO module creates a session id (sid) for each request.
    ' On connect the sid is saved in the connected sessions list.
    '''
    connected_sessions.append(request.sid)


@socketio.on('disconnect')
@login_required
def disconnect():
    '''
    ' On disconnect the session id gets removed from the connected sessions
    ' list.
    '''
    connected_sessions.remove(request.sid)


@socketio.on('subscribe_user_ressources')
@login_required
def subscribe_user_ressources():
    socketio.start_background_task(user_ressource_subscription_handler,
                                   current_app._get_current_object(),
                                   current_user.id, request.sid)


@socketio.on('subscribe_foreign_user_ressources')
@login_required
@admin_required
def subscribe_foreign_user_ressources(user_id):
    socketio.start_background_task(user_ressource_subscription_handler,
                                   current_app._get_current_object(),
                                   user_id, request.sid, True)


def user_ressource_subscription_handler(app, user_id, session_id,
                                        foreign=False):
    '''
    ' Sends initial corpus and job lists to the client. Afterwards it checks
    ' every 3 seconds if changes to the initial values appeared. If changes are
    ' detected, a RFC 6902 compliant JSON patch gets send.
    '
    ' NOTE: The initial values are send as a init-* events.
    '       The JSON patches are send as update-* events.
    '           > where '*' is either 'corpora' or 'jobs'
    '''
    logger = logging.getLogger(__name__)
    logger.warning('user_id: {}, foreign: {}'.format(user_id, foreign))
    init_events = {'corpora': 'init-foreign-corpora' if foreign
                   else 'init-corpora',
                   'jobs': 'init-foreign-jobs' if foreign else 'init-jobs'}
    update_events = {'corpora': 'update-foreign-corpora' if foreign
                     else 'update-corpora',
                     'jobs': 'update-foreign-jobs' if foreign
                     else 'update-jobs'}
    with app.app_context():
        # Gather current values from database.
        user = User.query.filter_by(id=user_id).first()
        corpora = user.corpora_as_dict()
        jobs = user.jobs_as_dict()
        # Send initial values to the user.
        socketio.emit(init_events['corpora'], json.dumps(corpora),
                      room=session_id)
        socketio.emit(init_events['jobs'], json.dumps(jobs), room=session_id)
        while session_id in connected_sessions:
            # Get new values from the database
            new_corpora = user.corpora_as_dict()
            new_jobs = user.jobs_as_dict()
            # Compute JSON patches.
            corpora_patch = jsonpatch.JsonPatch.from_diff(corpora, new_corpora)
            jobs_patch = jsonpatch.JsonPatch.from_diff(jobs, new_jobs)
            # In case there are patches, send them to the user.
            if corpora_patch:
                socketio.emit(update_events['corpora'],
                              corpora_patch.to_string(), room=session_id)
            if jobs_patch:
                socketio.emit(update_events['jobs'], jobs_patch.to_string(),
                              room=session_id)
            # Set new values as references for the next iteration.
            corpora = new_corpora
            jobs = new_jobs
            socketio.sleep(3)
